import math
import timeit

import numpy as np
from scipy.spatial import KDTree
from sklearn.decomposition import PCA

from .primitives.clustering import DBSCAN_clustering
from .primitives.voxel import voxelate

# -----------------------------------------------------------------------------
# compute_axes
# -----------------------------------------------------------------------------


def compute_axes_exact(
    voxelated_cloud,
    clust_stripe,
    stripe_lower_limit,
    stripe_upper_limit,
    h_range,
    min_points,
    d_max,
    X_field,
    Y_field,
    Z_field,
    Z0_field,
    tree_id_field,
    progress_hook=None,
):
    """Function used inside individualize_trees during tree individualization
    process. It identifies tree axes. It expects a voxelated version of the
    point cloud and a filtered (based on the verticality clustering process)
    stripe as input, so that it only contains (hopefully) stems. Those stems
    are isolated and enumerated, and then, their axes are identified using PCA
    (PCA1 direction). This allows to group points based on their distance to
    those axes, thus assigning each point to a tree.

    It requires a height-normalized cloud in order to function properly.

    Parameters
    ----------
    voxelated_cloud : numpy.ndarray
        The voxelated point cloud. It is expected to have X, Y, Z and Z0 fields.
    clust_stripe : numpy.ndarray
        The point cloud containing the clustered stripe (the stems) from
        verticality_clustering.
    stripe_lower_limit : float
        Lower (vertical) limit of the stripe (units is meters). Defaults to 0.7.
    stripe_upper_limit : float
        Upper (vertical) limit of the stripe (units is meters). Defaults to 3.5.
    h_range : float
        Only stems where points extend vertically throughout a range as tall as
        defined by h_range are considered.
    min_points : int
        Minimum number of points in a cluster for it to be considered as a
        potential stem
    d_max : float
        Points that are closer than d_max to an axis are assigned to that axis.
    X_field : int
        Index at which (x) coordinates are stored.
    Y_field : int
        Index at which (y) coordinates are stored.
    Z_field : int
        Index at which (z) coordinates are stored.
    Z0_field : int
        Index at which (z0) coordinates are stored.
    tree_id_field : int
        Index at which cluster ID is stored.
    progress_hook : callable, optional
        A hook that take two int, the first is the current number of iteration
        and the second is the targeted number iteration. Defaults to None.

    Returns
    -------
    detected_trees : numpy.ndarray
        Matrix with as many rows as trees, containing a description of each
        individualized tree. It stores the following values: tree ID, PCA1 X
        value, PCA1 Y value, PCA1 Z value, stem centroid X value, stem centroid
        Y value, stem centroid Z value, height difference of stem centroid
        (z - z0), axis vertical deviation.
    dist_to_axis : numpy.ndarray
        Matrix containing the distance from each point to the closest axis.
    tree_id_vector : numpy.ndarray
        Vector containing the tree IDs.
    """
    start_total = timeit.default_timer()
    # Empty vectors that will store final outputs: - distance from each point to
    # closest axis - ID of the corresponding tree (the tree that the point belongs to).
    dist_to_axis = np.zeros((np.size(voxelated_cloud, 0))) + 100000  # distance to the closest axis
    tree_id_vector = np.zeros((np.size(voxelated_cloud, 0))) + 100000  # tree ID of closest axis

    # Set of all possible trees (trunks at this stage) and number of points associated to each:
    unique_values, n = np.unique(clust_stripe[:, tree_id_field], return_counts=True)

    # Filtering of possible trees that do not contain enough points to be considered.
    filt_unique_values = unique_values[n > min_points]

    # Final number of trees
    n_values = np.size(filt_unique_values)

    # Empty array to be filled with several descriptors of the trees. In the following order:
    # tree ID | PCA1 X value | PCA1 Y value | PCA1 Z value | trunk centroid X value
    # | trunk centroid Y value | trunk centroid Z value | height difference |
    # It has as many rows as trees are.
    detected_trees = np.zeros((np.size(filt_unique_values, 0), 9))

    # Index used to display progress information.
    # we use this index because filt_unique_values contains tree_id that
    # could be non contiguous
    id_progress = 0
    # Index used to keep track of valid trees and index the detected_trees array.
    id_valid = 0

    # Height range (actual value, not the %) that points should extend throughout
    h_range_value = (stripe_upper_limit - stripe_lower_limit) * h_range

    # First loop: It goes over each tree (still stems) except for the first entry,
    # which maps to noise (this entry is generated by DBSCAN during clustering).
    if progress_hook is not None:
        progress_hook(0, n_values)
    for i in filt_unique_values:
        # Isolation of stems: stem_i only contains points associated to 1 tree.
        tree_mask = clust_stripe[:, tree_id_field] == i
        stem_i = clust_stripe[tree_mask][:, [X_field, Y_field, Z_field]]

        # Z and Z0 mean heights of points in a given tree
        z_z0 = np.mean(clust_stripe[tree_mask][:, [Z_field, Z0_field]], axis=0)

        # Difference between Z and Z0 mean heights
        diff_z_z0 = z_z0[0] - z_z0[1]

        # Second loop: only stems where points extend vertically throughout its
        # whole range are considered.
        if np.ptp(stem_i[:, Z_field]) > h_range_value:
            # PCA and centroid computation.
            pca_out = PCA(n_components=3)
            pca_out.fit(stem_i)
            centroid = np.mean(stem_i, axis=0)

            # Values are stored in tree vector
            detected_trees[id_valid, 0] = i  # tree ID
            detected_trees[id_valid, 1:4] = pca_out.components_[0, :]  # PCA1 X value | PCA1 Y value | PCA1 Z value
            detected_trees[id_valid, 4:7] = centroid  # Stem centroid coordinates
            detected_trees[id_valid, 7] = diff_z_z0  # Height difference
            detected_trees[id_valid, 8] = np.abs(
                np.degrees(
                    np.arctan(
                        np.hypot(detected_trees[id_valid, 1], detected_trees[id_valid, 2]) / detected_trees[id_valid, 3]
                    )
                )
            )

            # Coordinate transformation from original to PCA. Done for EVERY
            # point of the original cloud from the PCA of a SINGLE stem.
            cloud_pca_coords = pca_out.transform(voxelated_cloud[:, [X_field, Y_field, Z_field]])
            # Distance from every point in the new coordinate system to the axes.
            # It is directly computed from the quadratic component of PC2 and PC3
            axis_dist = np.hypot(cloud_pca_coords[:, 1], cloud_pca_coords[:, 2])
            # Points that are closer than d_max to an axis are assigned to that axis.
            # Also, if a point is closer to an axis than it was to previous axes, accounting for points
            # that were previously assigned to some other axis in previous iterations, it is assigned
            # to the new, closer axis. Distance to the axis is stored as well.
            valid_points = (axis_dist < d_max) & (axis_dist < dist_to_axis)
            tree_id_vector[valid_points] = i
            dist_to_axis[valid_points] = axis_dist[valid_points]
            # Progress hook
            id_valid = id_valid + 1
            id_progress = id_progress + 1
            if progress_hook is not None:
                progress_hook(id_progress, n_values)

        else:
            # we keep track of the progress even if the tree is not valid
            id_progress = id_progress + 1
            if progress_hook is not None:
                progress_hook(id_progress, n_values)
    print("-> Dist axes total time :", start_total - timeit.default_timer())
    # This deletes the trailing rows that only contains zeros
    detected_trees = detected_trees[~np.all(detected_trees == 0, axis=1)]
    return detected_trees, dist_to_axis, tree_id_vector


def compute_axes_approximate(
    voxelated_cloud,
    clust_stripe,
    stripe_lower_limit,
    stripe_upper_limit,
    h_range,
    min_points,
    voxel_resolution,
    d_max,
    X_field,
    Y_field,
    Z_field,
    Z0_field,
    tree_id_field,
    progress_hook=None,
):
    """Function used inside individualize_trees during tree individualization
    process. It identifies tree axes. It expects a voxelated version of the
    point cloud and a filtered (based on the verticality clustering process)
    stripe as input, so that it only contains (hopefully) stems. Those stems
    are isolated and enumerated, and then, their axes are identified using PCA
    (PCA1 direction). This allows to group points based on their distance to
    those axes, thus assigning each point to a tree.

    This is an approximate method that samples points along each axis and inserts
    them into a k-d tree.
    It enables finding the nearest neighbor (NN) point in the k-d tree index
    for each point in the voxelated cloud, allowing for parallel processing of
    an approximate point-line distances queries.

    It requires a height-normalized cloud in order to function properly.


    Parameters
    ----------
    voxelated_cloud : numpy.ndarray
        The voxelated point cloud. It is expected to have X, Y, Z and Z0 fields.
    clust_stripe : numpy.ndarray
        The point cloud containing the clustered stripe (the stems) from
        verticality_clustering.
    stripe_lower_limit : float
        Lower (vertical) limit of the stripe (units is meters). Defaults to 0.7.
    stripe_upper_limit : float
        Upper (vertical) limit of the stripe (units is meters). Defaults to 3.5.
    h_range : float
        Only stems where points extend vertically throughout a range as tall as
        defined by h_range are considered.
    min_points : int
        Minimum number of points in a cluster for it to be considered as a
        potential stem
    voxel_resolution: float
        Voxel resolution of the voxelated_cloud. It is used to compute the
        axis sampling step
    d_max : float
        Points that are closer than d_max to an axis are assigned to that axis.
    X_field : int
        Index at which (x) coordinates are stored.
    Y_field : int
        Index at which (y) coordinates are stored.
    Z_field : int
        Index at which (z) coordinates are stored.
    Z0_field : int
        Index at which (z0) coordinates are stored.
    tree_id_field : int
        Index at which cluster ID is stored.
    progress_hook : callable, optional
        A hook that take two int, the first is the current number of iteration
        and the second is the targeted number iteration. Defaults to None.

    Returns
    -------
    detected_trees : numpy.ndarray
        Matrix with as many rows as trees, containing a description of each
        individualized tree. It stores the following values: tree ID, PCA1 X
        value, PCA1 Y value, PCA1 Z value, stem centroid X value, stem centroid
        Y value, stem centroid Z value, height difference of stem centroid
        (z - z0), axis vertical deviation.
    dist_to_axis : numpy.ndarray
        Matrix containing the distance from each point to the closest axis.
    tree_id_vector : numpy.ndarray
        Vector containing the tree IDs.

    Raises
    ------
    RuntimeError
        If the axis is parallel to either the top or bottom plane, resulting in no intersection.
    """

    def _axis_bb_inter(axis_pos, axis_norm, tp_pos, bp_pos):
        """Calculate the intersection points of an axis with the top and bottom planes of a bounding box.

        This function computes the intersections of an axis, defined by a position and a normal vector,
        with two horizontal planes representing the top and bottom boundaries of a bounding box.

        We only test top and bottom for the sake of speed and simplicity but it could be overly conservative.

        Parameters
        ----------
        axis_pos : np.ndarray
            A 3D point representing a position on the axis.
        axis_norm : np.ndarray
            A 3D vector representing the normal vector of the axis.
        tp_pos : np.ndarray
            A 3D point representing a position on the top plane of the bounding box.
        bp_pos : np.ndarray
            A 3D point representing a position on the bottom plane of the bounding box.

        Returns
        -------
        bottom_limit: np.ndarray
            Intersection point with the bottom plane.
        top_limit: np.ndarray
            Intersection point with the top plane.

        Raises
        ------
        RuntimeError
            If the axis is parallel to either the top or bottom plane, resulting in no intersection.
        """
        top_limit = _vector_plane_intersection(axis_pos, axis_norm, tp_pos, np.array([0.0, 0.0, -1.0]))
        bottom_limit = _vector_plane_intersection(axis_pos, axis_norm, bp_pos, np.array([0.0, 0.0, 1.0]))

        # It should not be possible but we throw an exception one the axis
        # is parallel to one of the bounding plane
        # TODO: use maybe a more informative exception
        if top_limit is None or bottom_limit is None:
            raise RuntimeError("Tree axis can't be parallel to the top or the bottom plane of the bounding box")

        return bottom_limit, top_limit

    def _vector_plane_intersection(axis_pos, axis_norm, plane_pos, plane_norm):
        """
        Find the intersection point of a vector and a plane in 3D.

        Parameters
        ----------
        axis_pos : (numpy.ndarray)
            Starting 3D point of the vector.
        axis_norm : (numpy.ndarray)
            Normalized 3D (directional) vector of the axis.
        plane_pos : numpy.ndarray
            A 3D point on the plane.
        plane_norm : numpy.ndarray
            the 3D Normal vector of the plane.

        Returns
        -------
        numpy.array | None
            Intersection point or None if the vector is parallel to the plane.
        """
        EPS = 1.0e-06
        # Calculate the dot product axis_norm and the plane norm
        denom = np.dot(axis_norm, plane_norm)

        # no intersection, axis and plane are //
        if np.abs(denom) < EPS:
            return None

        # Projecting the vector from axis_pos to plane_pos onto the plane
        # normal, and then normalizing by denom.
        t = np.dot(plane_norm, plane_pos - axis_pos) / denom

        # Calculate the intersection point
        intersection = axis_pos + t * axis_norm

        return intersection

    start_total = timeit.default_timer()
    NO_ID = 100000  # ID for trees with dist_axis > d_max
    # Space between sample along the axes
    SAMPLE_STEP = voxel_resolution

    # Set of all possible trees (trunks at this stage) and number of points associated to each:
    unique_values, n = np.unique(clust_stripe[:, tree_id_field], return_counts=True)

    # Filtering of possible trees that do not contain enough points to be considered.
    filt_unique_values = unique_values[n > min_points]

    # Final number of trees
    n_values = np.size(filt_unique_values)

    # Empty array to be filled with several descriptors of the trees. In the following order:
    # It has as many rows as trees are.
    detected_trees = np.zeros((np.size(filt_unique_values, 0), 9))

    # Index used to display progress information.
    # we use this index because filt_unique_values contains tree_id that
    # could be non contiguous
    id_progress = 0
    # Index used to keep track of valid trees and index the detected_trees array.
    id_valid = 0

    # Height range (actual value, not the %) that points should extend throughout
    h_range_value = (stripe_upper_limit - stripe_lower_limit) * h_range

    # Compute bounding box of the voxelated PC
    bb_min = np.min(voxelated_cloud[:, [X_field, Y_field, Z_field]], axis=0)
    bb_max = np.max(voxelated_cloud[:, [X_field, Y_field, Z_field]], axis=0)

    axes_points_list = []
    valid_tree_ids = []

    # First loop: It goes over each tree (still stems) except for the first entry,
    # which maps to noise (this entry is generated by DBSCAN during clustering).
    if progress_hook is not None:
        progress_hook(0, n_values)
    for i in filt_unique_values:
        # Isolation of stems: stem_i only contains points associated to 1 tree.
        tree_mask = clust_stripe[:, tree_id_field] == i
        stem_i = clust_stripe[tree_mask][:, [X_field, Y_field, Z_field]]

        # Z and Z0 mean heights of points in a given tree
        z_z0 = np.mean(clust_stripe[tree_mask][:, [Z_field, Z0_field]], axis=0)

        # Difference between Z and Z0 mean heights
        diff_z_z0 = z_z0[0] - z_z0[1]

        # Second loop: only stems where points extend vertically throughout its
        # whole range are considered.
        if np.ptp(stem_i[:, Z_field]) > (h_range_value):
            # PCA and centroid computation.
            pca_out = PCA(n_components=3)
            pca_out.fit(stem_i)
            centroid = np.mean(stem_i, axis=0)

            # Values are stored in tree vector
            detected_trees[id_valid, 0] = i  # tree ID
            detected_trees[id_valid, 1:4] = pca_out.components_[0, :]  # PCA1 X value | PCA1 Y value | PCA1 Z value
            detected_trees[id_valid, 4:7] = (
                centroid  # stem centroid X value | stem centroid Y value | stem centroid Z value
            )
            detected_trees[id_valid, 7] = diff_z_z0  # Height difference
            detected_trees[id_valid, 8] = np.abs(
                np.degrees(
                    np.arctan(
                        np.hypot(detected_trees[id_valid, 1], detected_trees[id_valid, 2]) / detected_trees[id_valid, 3]
                    )
                )
            )

            min_point, max_point = _axis_bb_inter(centroid, detected_trees[id_valid, 1:4], bb_max, bb_min)

            distance = np.linalg.norm(max_point - min_point)
            num_samples = int(distance / SAMPLE_STEP)
            point_range = np.arange(0, num_samples)
            # upward vector
            vector = (
                -detected_trees[id_valid, 1:4] if detected_trees[id_valid, 3] < 0 else detected_trees[id_valid, 1:4]
            )
            axis_points = np.column_stack((point_range, point_range, point_range)) * vector * SAMPLE_STEP + min_point
            axes_points_list.append(axis_points)

            # append the current valid id for further reindexing
            valid_tree_ids.append(i)

            # Progress hook
            id_valid = id_valid + 1
            id_progress = id_progress + 1
            if progress_hook is not None:
                progress_hook(id_progress, n_values)
        else:
            # we keep track of the progress even if the tree is not valid
            id_progress = id_progress + 1
            if progress_hook is not None:
                progress_hook(id_progress, n_values)

    axes_cloud = np.concatenate(axes_points_list)

    # Compute distances and NN axis for each point via a KDTree.
    tree = KDTree(axes_cloud)
    dist_to_axis, indexes = tree.query(voxelated_cloud[:, [X_field, Y_field, Z_field]], 1, workers=-1)
    dist_to_axis[dist_to_axis > d_max] = d_max

    # fill tree_cluster
    tree_cluster = np.zeros(voxelated_cloud.shape[0])
    min_id = 0
    for i, axis_points in enumerate(axes_points_list):
        num_points = axis_points.shape[0]
        tree_cluster[min_id : min_id + num_points] = valid_tree_ids[i]
        min_id += num_points

    tree_id_vector = tree_cluster[indexes]
    tree_id_vector[dist_to_axis > d_max] = NO_ID
    print("-> Dist Axis total time :", timeit.default_timer() - start_total)

    # This deletes the trailing rows that only contains zeros
    detected_trees = detected_trees[~np.all(detected_trees == 0, axis=1)]
    return detected_trees, dist_to_axis, tree_id_vector


# -----------------------------------------------------------------------------
# compute_heights
# -----------------------------------------------------------------------------


def compute_heights(
    voxelated_cloud,
    detected_trees,
    dist_to_axis,
    tree_id_vector,
    d,
    max_dev,
    resolution_heights,
    n_digits,
    X_field,
    Y_field,
    Z_field,
):
    """Function used inside individualize_trees during tree individualization
    process. It measures tree heights. The function creates a large-resolution
    voxel cloud to and filters voxels containing few points. This has the
    purpose to discard any outlier point that might be over the trees, to then
    identify the highest point within the remaining voxels.

    It requires a height-normalized cloud in order to function properly.

    Parameters
    ----------
    voxelated_cloud : numpy.ndarray
        The voxelated point cloud. It is expected to have X, Y, Z and Z0 fields.
    detected_trees : numpy.ndarray
        See compute_axes.
    dist_to_axis : numpy.ndarray
        See compute_axes.
    tree_id_vector : numpy.ndarray
        See compute_axes.
    d : float
        Points within this distance from tree axis will be considered as
        potential points to define tree height.
    max_dev : float
        Maximum degree of vertical deviation of a tree axis to consider its
        tree height measurement as valid.
    resolution_heights : float
        (x, y, z) voxel resolution.
    n_digits : int
        Number of digits dedicated to each coordinate ((x), (y) or (z)) during
        the generation of each point code.
    X_field : int
        Index at which (x) coordinates are stored.
    Y_field : int
        Index at which (y) coordinates are stored.
    Z_field : int
        Index at which (z) coordinates are stored.

    Returns
    -------
    tree_heights: numpy.ndarray
        Matrix containing (x, y, z) coordinates of each tree's highest point,
        as well as its normalized height and a binary field stating if the
        axis was deviated (1) or if it was not (0).
    """

    # The cloud is re-voxelated to a larger resolution to then be clusterized.
    # Small clusters containing 1-2 voxels will be discarded to eliminate
    # outliers points that could interfere in height measurement.
    large_voxels_cloud, large_vox_to_cloud_ind, _ = voxelate(
        voxelated_cloud,
        resolution_heights,
        resolution_heights,
        n_digits,
        X_field,
        Y_field,
        Z_field,
        with_n_points=False,
    )

    # eps for DBSCAN
    eps_heights = resolution_heights * math.sqrt(3) + 1e-6

    # Large-resolution voxelated cloud is clusterized
    cluster_labels = DBSCAN_clustering(large_voxels_cloud, eps=eps_heights, min_samples=2)

    # Cluster labels are attached to the fine-resolution voxelated cloud
    voxelated_cloud = np.append(
        voxelated_cloud,
        np.expand_dims(cluster_labels[large_vox_to_cloud_ind], axis=1),
        axis=1,
    )

    # Tree IDS are attached to the fine-resolution voxelated cloud too
    voxelated_cloud = np.append(voxelated_cloud, np.expand_dims(tree_id_vector, axis=1), axis=1)

    # Eliminating all points too far away from axes
    voxelated_cloud = voxelated_cloud[dist_to_axis < d, :]

    # Set of all cluster labels and their cardinality: cluster_id = {1,...,K},
    # K = 'number of clusters'.
    cluster_id, K = np.unique(cluster_labels, return_counts=True)

    # Filtering of labels associated only to clusters that contain a minimum number of points.
    # Discarding points that do not belong to any cluster
    large_clusters = cluster_id[(K > 3) & (cluster_id != -1)]

    # Eliminating all points that belong to clusters with less than 2 points (large voxels)
    voxelated_cloud = voxelated_cloud[np.isin(voxelated_cloud[:, -2], large_clusters)]

    # Only pick the actual trees:
    n_trees = detected_trees.shape[0]
    tree_heights = np.zeros((n_trees, 5))

    for i in range(n_trees):  # Last row of tree_vector
        # Be aware this finds the highest voxel (fine-resolution), not the highest point.
        valid_id = detected_trees[i, 0]
        single_tree = voxelated_cloud[voxelated_cloud[:, -1] == valid_id, 0:3]  # Just the (x, y, z) coordinates
        which_z_max = np.argmax(single_tree[:, 2])  # The highest (z) value
        highest_point = single_tree[which_z_max, :]  # The highest point
        tree_heights[i, 0:3] = highest_point
        tree_heights[i, 3] = highest_point[2] - detected_trees[i, 7]  # (z0)

        # If tree is deviated from vertical, 1, else, 0.
        if detected_trees[i, -1] > [max_dev]:
            tree_heights[i, -1] = 0
        else:
            tree_heights[i, -1] = 1

    return tree_heights


# -----------------------------------------------------------------------------
# individualize_trees
# -----------------------------------------------------------------------------


def individualize_trees(
    cloud,
    clust_stripe,
    resolution_z=0.035,
    resolution_xy=0.035,
    stripe_lower_limit=0.7,
    stripe_upper_limit=3.5,
    h_range=1,
    d_max=1.5,
    min_points=20,
    d=15,
    max_dev=25,
    resolution_heights=0.3,
    n_digits=5,
    X_field=0,
    Y_field=1,
    Z_field=2,
    Z0_field=3,
    tree_id_field=-1,
    progress_hook=None,
):
    """This function expects a filtered (based on the clustering process)
    stripe as input, so that it only contains (hopefully) stems. Those stems
    are voxelated and enumerated, and then, their axes are identified using PCA
    (PCA1 direction). This allows to group points based on their distance to
    those axes, thus assigning each point to a tree. This last step is applied
    to the WHOLE original cloud. It also measures tree heights.

    It requires a height-normalized cloud in order to function properly.

    Parameters
    ----------
    cloud : numpy.ndarray
        The point cloud. It is expected to have X, Y, Z and Z0 fields.
    clust_stripe : numpy.ndarray.
        The point cloud containing the clusterized stripe from
        verticality_clustering. It is expected to have X, Y, Z0 and cluster ID
        fields.
    resolution_z : float
        (x, y) voxel resolution in meters. Defaults to 0.035.
    resolution_xy : float
        (z) voxel resolution in meters. Defaults to 0.035.
    stripe_lower_limit : float
        Lower (vertical) limit of the stripe (units is meters). Defaults to 0.7.
    stripe_upper_limit : float
        Upper (vertical) limit of the stripe (units is meters). Defaults to 3.5.
    h_range : float
        Only stems where points extend vertically throughout a range as tall as
        defined by h_range are considered (units is meters). Defaults to 1.2.
    d_max : float
        Points that are closer than d_max to an axis are assigned to that axis
        (units is meters). Defaults to 1.5.
    min_points : int
        Minimum number of points in a cluster for it to be considered as a
        potential stem. Defaults to 20.
    d : float
        Points within this distance from tree axis will be considered as
        potential points to define tree height (units is meters).
        Defaults to 15.
    max_dev : float
        Maximum degree of vertical deviation of a tree axis to consider its
        tree height measurement as valid (units is sexagesimal degrees).
        Defaults to 25.
    n_digits : int
        Number of digits dedicated to each coordinate ((x), (y) or (z))
        during the generation of each point code. Defaults to 5.
    resolution_heights : float
        (x, y, z) voxel resolution in meters used during height computation.
        Defaults to 0.3
    X_field : int
        Index at which (x) coordinate is stored. Defaults to 0.
    Y_field : int
        Index at which (y) coordinate is stored. Defaults to 1.
    Z_field : int
        Index at which (z) coordinate is stored. Defaults to 2.
    Z0_field : int
        Index at which (z0) coordinate is stored. Defaults to 3.
    tree_id_field : int
        Index at which cluster ID is stored. Defaults to -1.
    progress_hook : callable, optional
        A hook that take two int, the first is the current number of iteration
        and the second is the targeted number iteration. Defaults to None.

    Returns
    -------
    assigned_cloud : numpy.ndarray
        Point cloud containing individualized trees. It consists of
        (x), (y), (z) and (z0) coordinates, a 5th column containing tree ID
        that each point belongs to and a 6th column containing point distance
        to closest axis.
    detected_trees : numpy.ndarray
        Matrix with as many rows as trees, containing a description of each
        individualized tree. It stores tree ID, PCA1 X value, PCA1 Y value,
        PCA1 Z value, stem centroid X value, stem centroid Y value, stem
        centroid Z value, height difference of stem centroid (z - z0), axis
        vertical deviation.
    tree_heights : numpy.ndarray
        Matrix containing the heights of individualized trees. It consists of
        (x), (y), (z) and (z0) coordinates of the highest point of the tree and
        a 5th column containing a binary indicator: 0 - tree was too deviated
        from vertical, and height may not be accurate, or 1 - tree was not too
        deviated from vertical, thus height may be trusted.
    """

    # Whole original cloud voxelization
    voxelated_cloud, vox_to_cloud_ind, _ = voxelate(
        cloud,
        resolution_z,
        resolution_xy,
        n_digits,
        X_field,
        Y_field,
        Z_field,
        with_n_points=False,
    )

    # Call to compute_axes
    detected_trees, dist_to_axis, tree_id_vector = compute_axes_approximate(
        voxelated_cloud,
        clust_stripe,
        stripe_lower_limit,
        stripe_upper_limit,
        h_range,
        min_points,
        resolution_xy,
        d_max,
        X_field,
        Y_field,
        Z_field,
        Z0_field,
        tree_id_field,
        progress_hook,
    )

    # Call to compute_heights
    tree_heights = compute_heights(
        voxelated_cloud,
        detected_trees,
        dist_to_axis,
        tree_id_vector,
        d,
        max_dev,
        resolution_heights,
        n_digits,
        X_field,
        Y_field,
        Z_field,
    )

    # Two new fields are added to the original cloud: - tree ID (id of closest axis)
    # - distance to that axis
    assigned_cloud = np.hstack(
        [cloud, tree_id_vector[vox_to_cloud_ind, np.newaxis], dist_to_axis[vox_to_cloud_ind, np.newaxis]]
    )

    # Output: - Assigned cloud (X, Y, Z, Z0, tree_id, dist_to_axis) - tree vector
    return assigned_cloud, detected_trees, tree_heights



import math
import timeit

import numpy as np
import pgeof
from pgeof import EFeatureID

from .primitives.clustering import DBSCAN_clustering
from .primitives.voxel import voxelate

# -----------------------------------------------------------------------------
# verticality_clustering_iteration
# -----------------------------------------------------------------------------


def verticality_clustering_iteration(
    stripe,
    vert_scale,
    vert_threshold,
    n_points,
    resolution_xy,
    resolution_z,
    n_digits,
):
    """This function is to be used internally by verticality_clustering. The
    intended use of this function is to accept a stripe as an input, defined
    this as a subset of the original cloud delimited by a lower height and an
    upper height, which will narrow down a region where it is expected to only
    be stems. Then it will voxelate those points and compute the verticality
    via compute_features() from pgeof It will filter points based on
    their verticality value, voxelate again and then cluster the remaining
    points. Those are expected to belong to stems.

    Parameters
    ----------
    stripe : numpy.ndarray
        The point cloud containing the stripe. It is expected to have X, Y, Z0
        fields. 3D or higher array containing data with `float` type.
    vert_scale : float
        Scale to be used during verticality computation to define a
        neighborhood around a given point. Verticality will be computed from
        the structure tensor of said neighborhood via eigen-decomposition.
    vert_threshold : float
        Minimum verticality value associated to a point to consider it as part
        of a stem.
    n_points : int
        Minimum number of points in a cluster for it to be considered as a
        potential stem.
    resolution_xy : float
        (x, y) voxel resolution.
    resolution_z : float
        (z) voxel resolution.
    n_digits : int
        Number of digits dedicated to each coordinate ((x), (y) or (z)) during
        the generation of each point code.

    Returns
    -------
    clust_stripe: numpy.ndarray
        Point cloud containing the points from the stripe that are considered
        as stems. It consists of 4 columns: (x), (y) and (z) coordinates, and
        a 4th column containing the cluster ID of the cluster that each point
        belongs to.
    t1 : float
        Time spent.
    """

    t = timeit.default_timer()
    print(" -Computing verticality...")

    # Call to 'voxelate' function to voxelate the cloud.
    voxelated_stripe, vox_to_stripe_ind, _ = voxelate(
        stripe, resolution_xy, resolution_z, n_digits, with_n_points=False
    )
    # Computation of verticality values associated to voxels using
    # 'compute_features' function. It needs a vicinity radius, provided by
    # 'vert_scale'.
    # use a large max_knn like the one used by jakteristics (it could be lowered)
    vert_values = pgeof.compute_features_selected(voxelated_stripe, vert_scale, 50000, [EFeatureID.Verticality])

    elapsed = timeit.default_timer() - t
    print("   %.2f" % elapsed, "s")
    t1 = elapsed

    # Verticality values are appended to the ORIGINAL cloud, using voxel-to-
    # original-cloud indexes.
    vert_stripe = np.hstack((stripe, vert_values[vox_to_stripe_ind]))

    # Filtering of points that were in voxels whose verticality value is under
    # the threshold. Output is a filtered cloud.
    filt_stripe = vert_stripe[vert_stripe[:, -1] > vert_threshold]

    # Check there are enough points to continue
    if filt_stripe.shape[0] == 0:
        raise ValueError(
            "No vertical clusters where found with these parameters."
            "Suggestion: decrease n_points/voxel size or verticality "
            "threshold."
        )

    t = timeit.default_timer()
    print(" -Clustering...")

    # The filtered cloud is voxelated.
    vox_filt_stripe, vox_to_filt_stripe_ind, _ = voxelate(
        filt_stripe, resolution_xy, resolution_z, n_digits, with_n_points=False
    )

    eps = resolution_xy * math.sqrt(3) + 1e-6
    # Clusterization of the voxelated cloud obtained from the filtered cloud.
    cluster_labels = DBSCAN_clustering(vox_filt_stripe, eps=eps, min_samples=2)

    # Set of all cluster labels and their cardinality: cluster_id = {1,...,K},
    # K = 'number of clusters'.
    cluster_id, K = np.unique(cluster_labels, return_counts=True)

    # Raise error if there's only one cluster id (-1)
    if len(cluster_id) == 1 and cluster_id[0] == -1:
        raise ValueError(
            "No stems were found with the current configuration. Suggestion: decrease n_points/voxel size."
        )

    elapsed = timeit.default_timer() - t
    print("   %.2f" % elapsed, "s")
    t1 = elapsed + t1

    t = timeit.default_timer()
    print(" -Extracting 'candidate' stems...")

    # Cluster labels are appended to the FILTERED cloud. They map each point to
    # the cluster they belong to, according to the clustering algorithm.
    vox_filt_lab_stripe = np.append(
        filt_stripe,
        np.expand_dims(cluster_labels[vox_to_filt_stripe_ind], axis=1),
        axis=1,
    )

    # Filtering of labels associated only to clusters that contain a minimum
    # number of points.
    # Moreover, ID = -1 is always created by DBSCAN to include points
    # that were not included in any cluster.
    large_clusters = cluster_id[(K > n_points) & (cluster_id != -1)]

    # Raise error if there are no large clusters.
    if large_clusters.size == 0:
        raise ValueError(
            "Clusters were found, but they are too small to be considered potential "
            "stems using current settings. Suggestion: decrease n_points."
        )

    # Removing the points that are not in valid clusters.
    clust_stripe = vox_filt_lab_stripe[np.isin(vox_filt_lab_stripe[:, -1], large_clusters)]

    n_clusters = large_clusters.shape[0]

    elapsed = timeit.default_timer() - t
    print("   %.2f" % elapsed, "s")
    t1 = elapsed + t1
    print("   %.2f" % t1, "s per iteration")
    print("   ", n_clusters, " clusters")
    return clust_stripe, t1


# -----------------------------------------------------------------------------------------------------------------------------------
# verticality_clustering
# ----------------------------------------------------------------------------------------------------------------------------------------


def verticality_clustering(
    stripe,
    scale=0.1,
    vert_threshold=0.7,
    n_points=1000,
    n_iter=2,
    resolution_xy=0.02,
    resolution_z=0.02,
    n_digits=5,
):
    """This function implements a for loop that iteratively calls
    verticality_clustering_iteration, 'peeling off' the stems.

    Parameters
    ----------
    stripe : numpy.ndarray
        The point cloud containing the stripe. It is expected to have X, Y, Z0
        fields. 3D or higher array containing data with `float` type.
    scale : float
        Scale to be used during verticality computation to define a
        neighborhood around a given point. Verticality will be computed from
        the structure tensor of said neighborhood via Eigendecomposition.
        Defaults to 0.1.
    vert_threshold : float
        Minimum verticality value associated to a point to consider it as part
        of a stem. Defaults to 0.7.
    n_points : int
        Minimum number of points in a cluster for it to be considered as a
        potential stem. Defaults to 1000.
    n_iter : int
        Number of iterations of 'peeling'. Defaults to 2.
    resolution_xy : float
        (x, y) voxel resolution. Defaults to 0.02.
    resolution_z : float
        (z) voxel resolution. Defaults to 0.02.
    n_digits : int
        Number of digits dedicated to each coordinate ((x), (y) or (z)) during
        the generation of each point code. Defaults to 5.

    Returns
    -------
    clust_stripe : numpy.ndarray
        Point cloud containing the points from the stripe that are considered
        as stems. It consists of 4 columns: (x), (y) and (z) coordinates, and
        a 4th column containing the cluster ID of the cluster that each point
        belongs to.
    """

    # This first if loop is just a fix that allows to compute everything
    # ignoring verticality. It should be addressed as it currently computes
    # verticality when n_iter = 0 and that should not happen (although, in
    # practice, n_iter should never be 0). It does not provide wrong results
    # but it slows down the process needlessly.
    if n_iter == 0:
        n_iter = 1
        vert_threshold = 0

    # Basically, use verticality_clustering as many times as defined by n_iter
    aux_stripe = stripe
    total_t = 0
    for i in np.arange(n_iter):
        print("Iteration number", i + 1, "out of", n_iter)
        clust_stripe, t = verticality_clustering_iteration(
            aux_stripe, scale, vert_threshold, n_points, resolution_xy, resolution_z, n_digits
        )
        aux_stripe = clust_stripe
        total_t = total_t + t
    print("Final:")
    print("%.2f" % total_t, "s in total (whole process)")
    return clust_stripe



import numpy as np
from scipy import optimize as opt
from scipy.cluster import hierarchy as sch
from scipy.spatial import distance_matrix

# -----------------------------------------------------------------------------
# point_clustering
# -----------------------------------------------------------------------------


def point_clustering(X, Y, max_dist):
    """This function clusters points by distance and finds the largest
    cluster. It is to be used inside fit_circle_check().

    Parameters
    ----------
    X : numpy.ndarray
        Vector containing (x) coordinates of points belonging to a tree section.
    Y : numpy.ndarray
        Vector containing (y) coordinates of points belonging to a tree section.
    max_dist : float
        Max separation among the points to be considered as members of the same
        cluster.

    Returns
    -------
    X_g : numpy.ndarray
        Vector containing the (x) coordinates of the largest cluster.
    Y_g : numpy.ndarray
        Vector containing the (y) coordinates of the largest cluster.
    """

    # Stacks 1D arrays ([X], [Y]) into a 2D array ([X, Y])
    xy_stack = np.column_stack((X, Y))

    # sch.fclusterdata outputs a vector that contains cluster ID of each point
    # (which cluster does each point belong to)
    clust_id = sch.fclusterdata(xy_stack, max_dist, criterion="distance", metric="euclidean")

    # Set of all clusters
    clust_id_unique = np.unique(clust_id)

    # For loop that iterates over each cluster ID, sums its elements and finds
    # the largest
    n_max = 0
    for c in clust_id_unique:
        # How many elements are in each cluster
        n = np.sum(clust_id == c)

        # Update largest cluster and its cardinality
        if n > n_max:
            n_max = n
            largest_cluster = c

    # X, Y coordinates of points that belong to the largest cluster
    X_g = xy_stack[clust_id == largest_cluster, 0]
    Y_g = xy_stack[clust_id == largest_cluster, 1]

    # Output: those X, Y coordinates
    return X_g, Y_g


# -------------------------------------------------------------------------------------------------------------------------------------------------------
# fit_circle
# -------------------------------------------------------------------------------------------------------------------------------------------------------


def fit_circle(X, Y):
    """This function fits points within a tree section into a circle by
    least squares minimization. It is to be used inside fit_circle_check().

    Parameters
    ----------
    X : numpy.ndarray
        Vector containing (x) coordinates of points belonging to a tree section.
    Y : numpy.ndarray
        Vector containing (y) coordinates of points belonging to a tree section.

    Returns
    -------
    circle_c : numpy.ndarray
        Matrix containing the (x, y) coordinates of the circle center.
    mean_radius : numpy.ndarray
        Vector containing the radius of each fitted circle
        (units is meters).
    """

    # Function that computes distance from each 2D point to a single point defined by (X_c, Y_c)
    # It will be used to compute the distance from each point to the circle center.
    def _calc_R(X, Y, X_c, Y_c):
        return np.sqrt((X - X_c) ** 2 + (Y - Y_c) ** 2)

    # Function that computes algebraic distance from each 2D point to some middle circle c
    # It calls calc_R (just defined above) and it is used during the least squares optimization.
    def _f_2(c, X, Y):
        R_i = _calc_R(X, Y, *c)
        return R_i - R_i.mean()

    # Initial barycenter coordinates (middle circle c center)
    X_m = X.mean()
    Y_m = Y.mean()
    barycenter = X_m, Y_m

    # Least square minimization to find the circle that best fits all
    # points within the section. 'ier' is a flag indicating whether the solution
    # was found (ier = 1, 2, 3 or 4) or not (otherwise).
    circle_c, _ = opt.leastsq(_f_2, barycenter, args=(X, Y), maxfev=2000)

    # Its radius
    radius = _calc_R(X, Y, *circle_c)
    mean_radius = radius.mean()

    # Output: - X, Y coordinates of best-fit circle center - its radius
    return circle_c, mean_radius


# -----------------------------------------------------------------------------
# inner_circle
# -----------------------------------------------------------------------------


def inner_circle(X, Y, X_c, Y_c, R, times_R):
    """Function that computes an internal circle inside the one fitted by
    fit_circle. This new circle is used as a validation tool and it gives
    insight on the quality of the 'fit_circle-circle'.

        - If points are closest to the inner circle, then the first fit was not
          appropriate

        - On the contrary, if points are closer to the outer circle, the
          'fit_circle-circle' is appropriate and describes well the stem diameter.

    Instead of directly computing the inner circle, it just takes a proportion
    (less than one) of the original circle radius and its center. Then, it just
    checks how many points are closest to the inner circle than to the original
    circle.

    Parameters
    ----------
    X : numpy.ndarray
        Vector containing (x) coordinates of points belonging to a tree section.
    Y : numpy.ndarray
        Vector containing (y) coordinates of points belonging to a tree section.
    X_c : numpy.ndarray
    Vector containing (x) coordinates of fitted circles.
    Y_c : numpy.ndarray
        Vector containing (y) coordinates of fitted circles.
    R : numpy.ndarray
        Vector containing the radii of the fitted circles.

    Returns
    -------
    n_points_in : numpy.ndarray
        Vector containing the number of points inside the inner circle of each
        section.
    """

    # Distance from each 2D point to the center.
    distance = np.sqrt((X - X_c) ** 2 + (Y - Y_c) ** 2)

    # Number of points closest to the inner circle, whose radius is
    # proportionate to the outer circle radius by a factor defined by 'times_R'.
    n_points_in = np.sum(distance < R * times_R)

    # Output: Number of points closest to the inner circle.
    return n_points_in


# -------------------------------------------------------------------------------------------------------------------------------------------------------
# sector_occupancy
# -------------------------------------------------------------------------------------------------------------------------------------------------------


def sector_occupancy(X, Y, X_c, Y_c, R, n_sectors, min_n_sectors, width):
    """This function provides quality measurements for the fitting of the
    circle. It divides the section in a number of sectors to check if there are
    points within them (so they are occupied). If there are not enough occupied
    sectors, the section fails the test, as it is safe to assume it has an
    abnormal, non desirable structure.

    Parameters
    ----------
    X : numpy.ndarray
        Vector containing (x) coordinates of points belonging to a tree section.
    Y : numpy.ndarray
        Vector containing (y) coordinates of points belonging to a tree section.
    X_c : numpy.ndarray
        Vector containing (x) coordinates of fitted circles.
    Y_c : numpy.ndarray
        Vector containing (y) coordinates of fitted circles.
    R : numpy.ndarray
        Vector containing the radii of the fitted circles.
    n_sectors : int
        Number of sectors in which sections will be divided.
    min_n_sectors : int
        Minimum number of occupied sectors in a section for its fitted circle
        to be considered as valid.
    width : float
        Width around the fitted circle to look for points (units is
        meters).

    Returns
    -------
    perct_occupied_sectors : float
        Percentage of occupied sectors in each section.
    enough_occupied_sectors : int
        Binary indicators whether the fitted circle is valid
        or not. 1 - valid, 0 - not valid.
    """

    # Coordinates translation.
    X_red = X - X_c
    Y_red = Y - Y_c

    # Computation of radius and angle necessary to transform cartesian coordinates
    # to polar coordinates.
    radial_coord = np.sqrt(X_red**2 + Y_red**2)  # radial coordinate
    angular_coord = np.arctan2(X_red, Y_red)  # angular coordinate. This function from numpy directly computes it.

    # Points that are close enough to the circle that will be checked.
    points_within = (radial_coord > (R - width)) * (radial_coord < (R + width))

    # Codification of points in each sector. Basically the range of angular coordinates
    # is divided in n_sector pieces and granted an integer number. Then, every
    # point is assigned the integer corresponding to the sector it belongs to.
    norm_angles = np.floor(
        angular_coord[points_within] / (2 * np.pi / n_sectors)
    )  # np.floor only keep the integer part of the division

    # Number of points in each sector.
    n_occupied_sectors = np.size(np.unique(norm_angles))

    # Percentage of occupied sectors.
    perct_occupied_sectors = n_occupied_sectors * 100 / n_sectors

    # If there are enough occupied sectors, then it is a valid section.
    enough_occupied_sectors = 0 if n_occupied_sectors < min_n_sectors else 1  # TODO(RJ): Maybe convert this to boolean

    # Output: percentage of occupied sectors | boolean indicating if it has enough
    # occupied sectors to pass the test.
    return perct_occupied_sectors, enough_occupied_sectors


# -----------------------------------------------------------------------------
# fit_circle_check
# -----------------------------------------------------------------------------


def fit_circle_check(
    X,
    Y,
    review,
    second_time,
    times_R,
    threshold,
    R_min,
    R_max,
    max_dist,
    n_points_section,
    n_sectors,
    min_n_sectors,
    width,
):
    """This function calls fit_circle() to fit points within a section to a
    circle by least squares minimization. These circles will define tree
    sections. It checks the goodness of fit using sector_occupancy and
    inner_circle. If fit is not appropriate, another circle will be fitted
    using only points from the largest cluster inside the first circle.

    Parameters
    ----------
    X : numpy.ndarray
        Vector containing (x) coordinates of points belonging to a tree section.
    Y : numpy.ndarray
        Vector containing (y) coordinates of points belonging to a tree section.
    second_time : int
        Integer that indicates whether it is the first time
        a circle is fitted or not (will be modified internally).
    times_R : float
        Ratio of radius between outer circle and inner circle.
    threshold : float
        Minimum number of points in inner circle for a fitted circle to be
        valid.
    R_min : float
        Minimum radius that a fitted circle must have to be valid.
    R_max : float
        Maximum radius that a fitted circle must have to be valid.
    max_dist : float
        Max separation among the points to be considered as members of the same
        cluster.
    n_points_section : int
        Minimum points within a section for its fitted circle to be valid.
    n_sectors : int
        Number of sectors in which sections will be divided.
    min_n_sectors : int
        Minimum number of occupied sectors in a section for its fitted circle
        to be considered as valid.
    width : float
        Width around the fitted circle to look for points (units is millimeters).

    Returns
    -------
    X_gs : numpy.ndarray
        Matrix containing (x) coordinates of largest clusters.
    Y_gs : numpy.ndarray
        Matrix containing (y) coordinates of largest clusters.
    X_c : numpy.ndarray
        Matrix containing (x) coordinates of the center of the best-fit circles.
    Y_c : numpy.ndarray
        Matrix containing (y) coordinates of the center of the best-fit circles.
    R : numpy.ndarray
        Vector containing best-fit circle radii.
    section_perct : numpy.ndarray
        Matrix containing the percentage of occupied sectors.
    n_points_in : numpy.ndarray
        Matrix containing the number of points in the inner circle.
    """

    # If loop that discards sections that do not have enough points (n_points_section)
    if X.size > n_points_section:
        # Call to fit_circle to fit the circle that best fits all points
        # within the section.
        (circle_center, R) = fit_circle(X=X, Y=Y)
        X_c = circle_center[0]  # Column 0 is center X coordinate
        Y_c = circle_center[1]  # Column 1 is center Y coordinate

        # Call to inner_circle to fit an inner circle and to get the number
        # of points closest to it.
        n_points_in = inner_circle(X, Y, X_c, Y_c, R, times_R)

        # Call to sector_occupancy to check if sectors around inner circle are occupied.
        sector_perct, enough_sectors = sector_occupancy(X, Y, X_c, Y_c, R, n_sectors, min_n_sectors, width)

        # If any of the following conditions hold:
        #   - Too many points in inner circle
        #   - Radius of best-fit circle is too small
        #   - Number of occupied sectors is too low
        # Then proceed with countermeasures
        if n_points_in > threshold or R < R_min or R > R_max or enough_sectors == 0:
            # If this is not the second round or, simply, if it is the first round,
            # then proceed
            if second_time == 0:
                # First round implies there is no X_g or Y_g, as points would not
                # have been grouped yet. point_clustering is called.
                X_g, Y_g = point_clustering(X, Y, max_dist)  # X_g or Y_g are the coordinates of the largest cluster.

                # If cluster size is big enough, then proceed. It is done this way to
                # account for cases where, even though the section had enough points,
                # there might not be enough points within the largest cluster.
                if X_g.size > n_points_section:
                    # Call to fit_circle_check (lets call it the 'deep call').
                    # Now it is guaranteed that it is a valid section (has enough
                    # points and largest cluster has enough points as well).
                    (
                        X_c,
                        Y_c,
                        R,
                        review,
                        second_time,
                        sector_perct,
                        n_points_in,
                    ) = fit_circle_check(
                        X_g,
                        Y_g,
                        0,
                        1,
                        times_R,
                        threshold,
                        R_min,
                        R_max,
                        max_dist,
                        n_points_section,
                        n_sectors,
                        min_n_sectors,
                        width,
                    )

                # If cluster size is not big enough, then don't take the section
                # it belongs to into account.
                else:
                    review = 1  # Even if it is not a valid section, lets note it has been checked.
                    X_c = 0
                    Y_c = 0
                    R = 0
                    second_time = 1

            # If this is the second round (whether the first round successfully
            # provided a valid section or not), then proceed.
            else:
                review = 1  # Just stating that if this is the second round, the check has happened.

    # This matches the first loop. If section is not even big enough (does not contain enough points), it is not valid.
    else:
        review = 2
        X_c = 0
        Y_c = 0
        R = 0
        second_time = 2
        sector_perct = 0
        n_points_in = 0

    return X_c, Y_c, R, review, second_time, sector_perct, n_points_in


# -----------------------------------------------------------------------------
# compute_sections
# -----------------------------------------------------------------------------


def compute_sections(
    stems,
    sections,
    section_width=0.02,
    times_R=0.5,
    threshold=5,
    R_min=0.03,
    R_max=0.5,
    max_dist=0.02,
    n_points_section=80,
    n_sectors=16,
    min_n_sectors=9,
    width=2,
    X_field=0,
    Y_field=1,
    Z0_field=3,
    tree_id_field=4,
    progress_hook=None,
):
    """This function calls fit_circle_check() to compute stem diameter at
    given sections.

    Parameters
    ----------
    stems : numpy.ndarray
        Point cloud containing the individualized trees. It is expected to
        have X, Y, Z0 and tree_ID fields.
    sections : numpy.ndarray
        Matrix containing a range of height values at which sections will be
        computed.
    section_width : float
        Points within this distance from any `sections` value will be considered
        as belonging to said section (units is meters). Defaults to 0.02.
    times_R : float
        Refer to fit_circle_check. Defaults to 0.5.
    threshold : float
        Refer to fit_circle_check. Defaults to 5.
    R_min : float
        Refer to fit_circle_check. Defaults to 0.03.
    R_max : float
        Refer to fit_circle_check. Defaults to 0.5.
    max_dist : float
        Refer to fit_circle_check. Defaults to 0.02.
    n_points_section : int
        Refer to fit_circle_check. Defaults to 80.
    n_sectors : int
        Refer to fit_circle_check. Defaults to 16.
    min_n_sectors : int
        Refer to fit_circle_check. Defaults to 9.
    width : float
        Refer to fit_circle_check. Defaults to 2.0.
    X_field : int
        Index at which (x) coordinate is stored. Defaults to 0.
    Y_field : int
        Index at which (y) coordinate is stored. Defaults to 1.
    Z0_field : int
        Index at which (z0) coordinate is stored. Defaults to 3.
    tree_id_field : int
        Index at which cluster ID is stored. Defaults to 4.
    progress_hook : callable, optional
        A hook that take two int, the first is the current number of iteration
        and the second is the targeted number iteration. Defaults to None.

    Returns
    -------
    X_c : numpy.ndarray
        Matrix containing (x) coordinates of the center of the best-fit circles.
    Y_c : numpy.ndarray
        Matrix containing (y) coordinates of the center of the best-fit circles.
    R : numpy.ndarray
        Vector containing best-fit circle radii.
    section_perct : numpy.ndarray
        Matrix containing the percentage of occupied sectors.
    n_points_in : numpy.ndarray
        Matrix containing the number of points in the inner circles.
    """
    trees = np.unique(stems[:, tree_id_field])  # Select the column that contains tree ID
    n_trees = trees.size  # Number of trees
    n_sections = sections.size  # Number of sections

    X_c = np.zeros((n_trees, n_sections))  # Empty array to store X data
    Y_c = np.zeros((n_trees, n_sections))  # Empty array to store Y data
    R = np.zeros((n_trees, n_sections))  # Empty array to store radius data
    check_circle = np.zeros((n_trees, n_sections))  # Empty array to store 'check' data
    second_time = np.zeros((n_trees, n_sections))  # Empty array to store 'second_time' data
    sector_perct = np.zeros((n_trees, n_sections))  # Empty array to store percentage of occupied sectors data
    n_points_in = np.zeros((n_trees, n_sections))  # Empty array to store inner points data

    # Filling previous empty arrays

    # Auxiliary index for first loop
    tree = -1  # Loop will start at -1
    if progress_hook is not None:
        progress_hook(0, n_trees)
    # First loop: iterates over each tree
    for tr in trees:
        # Tree ID is used to iterate over trees
        tree_i = stems[stems[:, tree_id_field] == tr, :]
        tree = tree + 1
        if progress_hook is not None:
            progress_hook(tree + 1, n_trees)
        # Auxiliary index for second loop
        section = 0

        # Second loop: iterates over each section
        for b in sections:
            # Selecting (x, y) coordinates of points within the section
            X = tree_i[
                (tree_i[:, Z0_field] >= b) & (tree_i[:, Z0_field] < b + section_width),
                X_field,
            ]
            Y = tree_i[
                (tree_i[:, Z0_field] >= b) & (tree_i[:, Z0_field] < b + section_width),
                Y_field,
            ]

            # fit_circle_check call. It provides data to fill the empty arrays
            (
                X_c[tree, section],
                Y_c[tree, section],
                R[tree, section],
                check_circle[tree, section],
                second_time[tree, section],
                sector_perct[tree, section],
                n_points_in[tree, section],
            ) = fit_circle_check(
                X,
                Y,
                0,
                0,
                times_R,
                threshold,
                R_min,
                R_max,
                max_dist,
                n_points_section,
                n_sectors,
                min_n_sectors,
                width,
            )

            section = section + 1
    return (X_c, Y_c, R, check_circle, second_time, sector_perct, n_points_in)


# -----------------------------------------------------------------------------
# tilt_detection
# -----------------------------------------------------------------------------


def tilt_detection(X_tree, Y_tree, radius, sections, Z_field=2, w_1=3.0, w_2=1.0):
    """This function finds outlier tilting values among sections within a tree
    and assigns a score to the sections based on those outliers. Two kinds of
    outliers are considered.

        - Absolute outliers are obtained from the sum of the deviations from
          every section center to all axes within a tree (the most tilted sections
          relative to all axes)

        - Relative outliers are obtained from the deviations of other section
          centers from a certain axis, within a tree (the most tilted sections
          relative to a certain axis)

    The 'outlier score' consists on a weighted sum of the absolute tilting value
    and the relative tilting value.

    Parameters
    ----------
    X_tree : numpy.ndarray
        Matrix containing (x) coordinates of the center of the sections.
    Y_tree : numpy.ndarray
        Matrix containing (y) coordinates of the center of the sections.
    radius : numpy.ndarray
        Vector containing section radii.
    sections : numpy.ndarray
        Vector containing the height of the section associated to each section.
    Z_field : int
        Index at which (z) coordinate is stored. Defaults to 2.
    w_1 : float
        Weight of absolute deviation. Defaults to 3.0.
    w_2 : float
        Weight of relative deviation. Defaults to 1.0.

    Returns
    -------
    outlier_prob : numpy.ndarray
        Vector containing the 'outlier probability' of each section.
    """

    # This function simply defines 1st and 3rd quartile of a vector and separates
    # values that are outside the interquartile range defined by these. Those
    # are the candidates to be outliers. This filtering may be done either
    # directly from the interquartile range, or from a certain distance from it,
    # thanks to 'n_range' parameter. Its default value is 1.5.

    def _outlier_vector(vector, lower_q=0.25, upper_q=0.75, n_range=1.5):
        q1, q3 = np.quantile(vector, [lower_q, upper_q])  # First quartile and Third quartile
        iqr = q3 - q1  # Interquartile range

        lower_bound = q1 - iqr * n_range  # Lower bound of filter. If n_range = 0 -> lower_bound = q1
        upper_bound = q3 + iqr * n_range  # Upper bound of filter. If n_range = 0 -> upper_bound = q3

        # return the outlier vector.
        return ((vector < lower_bound) | (vector > upper_bound)).astype(int)

    # Empty matrix that will store the probabilities of a section to be invalid
    outlier_prob = np.zeros_like(X_tree)

    # First loop: iterates over each tree
    for i in range(X_tree.shape[0]):
        # If there is, at least, 1 circle with positive radius in a tree, then
        # proceed (invalid circles are stored with a radius value of 0)
        if np.sum(radius[i, :]) > 0:
            # Filtering sections within a tree that have valid circles (non-zero radius).
            valid_radius = radius[i, :] > 0
            num_valid_sections = np.size(sections[valid_radius])
            # Weights associated to each section. They are computed in a way
            # that the final value of outliers sums up to 1 as maximum.
            abs_outlier_w = w_1 / (num_valid_sections * w_2 + w_1)
            rel_outlier_w = w_2 / (num_valid_sections * w_2 + w_1)

            # Vertical distance matrix among all sections (among their centers)
            # Empty matrix to store heights of each section
            heights = np.zeros((num_valid_sections, Z_field))
            #  Height (Z value) of each section
            heights[:, 0] = np.transpose(sections[valid_radius])
            # Vertical distance matrix
            z_dist_matrix = distance_matrix(heights, heights)

            # Horizontal distance matrix among all sections (among their centers)
            # Store X, Y coordinates of each section
            c_coord = np.column_stack((X_tree[i][valid_radius], Y_tree[i][valid_radius]))
            # Horizontal distance matrix
            xy_dist_matrix = distance_matrix(c_coord, c_coord)

            # Tilting measured from every vertical within a tree: All verticals
            # obtained from the set of sections within a tree. For instance, if
            # there are 10 sections, there are 10 tilting values for each section.
            tilt_matrix = np.degrees(np.arctan(xy_dist_matrix / z_dist_matrix))

            # Summation of tilting values from each center.
            tilt_sum = np.nansum(tilt_matrix, axis=0)

            # Outliers within previous vector (too low / too high tilting values).
            # These are abnormals tilting values from ANY axis.
            outlier_prob[i][valid_radius] = _outlier_vector(tilt_sum) * abs_outlier_w

            # Second loop: iterates over each section (within a single tree).
            for j in range(np.size(sections[valid_radius])):
                # Search for abnormals tilting values from a CERTAIN axis.
                tilt_matrix[j, j] = np.quantile(tilt_matrix[j, ~j], 0.5)
                # Storing those values.
                rel_outlier = _outlier_vector(tilt_matrix[j]) * rel_outlier_w
                # Sum of absolute outlier value and relative outlier values
                outlier_prob[i][valid_radius] += rel_outlier

    return outlier_prob


# -----------------------------------------------------------------------------
# tree_locator
# --------------------------------------------------------------------------


def tree_locator(
    sections,
    X_c,
    Y_c,
    tree_vector,
    sector_perct,
    R,
    outliers,
    X_field=0,
    Y_field=1,
    Z_field=2,
):
    """This function generates points that locate the individualized trees and
    computes their DBH (diameter at breast height). It uses all the quality
    measurements defined in previous functions to check whether the DBH should
    be computed or not and to check which point should be used as the tree locator.

    The tree locators are then saved in a LAS file. Each tree locator corresponds
    on a one-to-one basis to the individualized trees.

    Parameters
    ----------
    sections : numpy.ndarray
        Vector containing section heights (normalized heights).
    X_c : numpy.ndarray
        Matrix containing (x) coordinates of the center of the sections.
    Y_c : numpy.ndarray
        Matrix containing (y) coordinates of the center of the sections.
    tree_vector : numpy.ndarray
        detected_trees output from individualize_trees.
    sector_perct : numpy.ndarray
        Matrix containing the percentage of occupied sectors.
    R : numpy.ndarray
        Vector containing section radii.
    outliers : numpy.ndarray
        Vector containing the 'outlier probability' of each section.
    X_field : int
        Index at which (x) coordinate is stored. Defaults to 0.
    Y_field : int
        Index at which (y) coordinate is stored. Defaults to 1.
    Z_field : int
        Index at which (z) coordinate is stored. Defaults to 2.

    Returns
    -------
    dbh_values : numpy.ndarray
        Vector containing DBH values.
    tree_locations : numpy.ndarray
        Matrix containing (x, y, z) coordinates of each tree locator.
    """
    DBH = 1.3  # Breast height constant

    # Number of trees
    n_trees = X_c.shape[0]
    # Empty vector to be filled with tree locators
    tree_locations = np.zeros((n_trees, 3))
    # Empty vector to be filled with DBH values.
    dbh_values = np.zeros((n_trees, 1))

    def _axis_location(index):
        """Given an index compute tree location from axis"""
        vector = -tree_vector[index, 1:4] if tree_vector[index, 3] < 0 else tree_vector[index, 1:4]
        dbh_values[index] = 0
        # Compute the height difference between centroid and BH
        diff_height = DBH - tree_vector[index, 6] + tree_vector[index, 7]
        # Compute the distance between centroid and axis point at BH.
        dist_centroid_dbh = diff_height / np.cos(np.radians(tree_vector[index, 8]))
        # Compute coordinates of axis point at BH.
        tree_locations[i, :] = vector * dist_centroid_dbh + tree_vector[i, 4:7]

    def _dbh_location(index, which_dbh):
        """Given an index, compute the tree location from the computed DBH"""
        dbh_values[index] = R[index, which_dbh] * 2
        # Their centers are averaged and we keep that value
        tree_locations[index, X_field] = X_c[index, which_dbh]
        tree_locations[index, Y_field] = Y_c[index, which_dbh]
        # Original height is obtained
        tree_locations[index, Z_field] = tree_vector[index, 7] + DBH

    # This if loop covers the cases where the stripe was defined in a way that
    # it did not include BH and DBH nor tree locator cannot be obtained from a
    # section at or close to BH. If that happens, tree axis is used to locate
    # the tree and DBH is not computed.
    if np.min(sections) > DBH:
        for i in range(n_trees):
            _axis_location(i)
    else:
        d = 1
        which_dbh = np.argmin(np.abs(sections - DBH))  # Which section is closer to BH.

        # get surrounding sections too
        lower_d_section = max(0, which_dbh - d)
        upper_d_section = min(sections.shape[0], which_dbh + d)
        # BH section and its neighbors. From now on, neighborhood
        close_to_dbh = np.arange(lower_d_section, upper_d_section + 1)  # upper bound is exclusive

        for i in range(n_trees):  # For each tree
            which_valid_R = R[i, close_to_dbh] > 0  # From neighborhood, select only those with non 0 radius
            # From neighborhood, select only those with outlier probability lower than 30 %
            which_valid_out = outliers[i, close_to_dbh] < 0.3
            # only those with sector occupancy higher than 30 %
            which_valid_sector_perct = sector_perct[i, close_to_dbh] > 30.0
            # valid points could be retrieved as well / i.e. only those with enough points in inner circle

            # If there are valid sections among the selected
            if np.any(which_valid_R) & np.any(which_valid_out):
                # If first section is BH section and if itself and its only neighbor are valid
                if (
                    (lower_d_section == 0)
                    & (np.all(which_valid_R))
                    & (np.all(which_valid_out))
                    & np.all(which_valid_sector_perct)
                ):  # Only happens when which_dbh == 0 in this case which_valid_points should be used here
                    # If they are coherent: difference among their radii is not larger than 10 % of the largest radius
                    if np.abs(R[i, close_to_dbh[0]] - R[i, close_to_dbh[1]]) < np.max(R[i, close_to_dbh]) * 0.1:
                        _dbh_location(i, which_dbh)
                    # If not all of them are valid, then there is no coherence and the axis location is used
                    else:
                        _axis_location(i)

                # If last section is BH section and if itself and its only neighbor are valid
                elif (upper_d_section == sections.shape[0]) & (np.all(which_valid_R)) & (np.all(which_valid_out)):
                    # if they are coherent; difference among their radii is not larger than 15 % of the largest radius
                    if np.abs(R[i, close_to_dbh[0]] - R[i, close_to_dbh[1]]) < np.max(R[i, close_to_dbh]) * 0.15:
                        # use BH section diameter as DBH
                        _dbh_location(i, which_dbh)

                    # If not all of them are valid, then there is no coherence in
                    # any case, and the axis location is used and DBH is not computed
                    else:
                        _axis_location(i)

                # In any other case, BH section is not first or last section, so it has 2 neighbors
                # 3 possibilities left:
                # A: Not all of three sections are valid: there is no possible coherence
                # B: All of three sections are valid, and there is coherence among the three
                # C: All of three sections are valid, but there is only coherence among neighbors
                # and not BH section or All of three sections are valid, but there is no coherence
                else:
                    # Case A:
                    if not ((np.all(which_valid_R)) & (np.all(which_valid_out)) & np.all(which_valid_sector_perct)):
                        _axis_location(i)
                    # case B&C:
                    else:
                        valid_sections = close_to_dbh  # Valid sections indexes
                        valid_radii = R[i, valid_sections]  # Valid sections radii
                        median_radius = np.median(valid_radii)  # Valid sections median radius
                        # Valid sections absolute deviation from median radius
                        abs_dev = np.abs(valid_radii - median_radius)
                        mad = np.median(abs_dev)  # Median absolute deviation
                        # Only keep sections close to median radius (3 MAD criterion)
                        filtered_sections = valid_sections[abs_dev < 3 * mad]
                        # 3 things can happen here:
                        # There are no deviated sections --> there is coherence among 3 --> case B
                        # There are 2 deviated sections --> only median radius survives filter --> case C
                        # Case B
                        if filtered_sections.shape[0] == close_to_dbh.shape[0]:
                            _dbh_location(i, which_dbh)
                        # Case C
                        else:
                            _axis_location(i)
            # If there is not a single section that either has non 0 radius nor low
            # outlier probability, there is nothing else to do -> axis location is used
            else:
                _axis_location(i)

    return dbh_values, tree_locations



from sklearn.cluster import DBSCAN


def DBSCAN_clustering(xyz, eps, min_samples):
    """
    Cluster 3D dimensional point clouds.

    It's essentially a wrapper around scikit-learn's DBSCAN.
    If `dendroptimized` is installed, an optimized version is used instead.
    However, this optimized version has an additional constraint:
    the point cloud must be regularly sampled, and `eps` must equal
    the sampling resolution of the point cloud x sqrt(3) i.e. the diagonal
    of a voxel (the enclosing ball of the voxel).

    Parameters
    ----------
    xyz : numpy.ndarray
        The point cloud to be denoised. Matrix containing (x, y, z) coordinates
        of the points.

    eps : float
        The clustering radius.

    min_samples : int
        The number of neighbors for a point to be considered as a core point.
        This includes the point itself.

    Returns
    -------
    labels : numpy.ndarray
        Cluster labels for each point. Noisy samples are labeled "-1".
    """
    return DBSCAN(eps=eps, min_samples=min_samples, n_jobs=-1).fit(xyz).labels_


# monkey patch clustering if possible
try:
    import dendroptimized

    DBSCAN_clustering = dendroptimized.connected_components
except ImportError:
    pass



import numpy as np

import dendromatics as dm


def test_clean_ground():
    # Regular grid of points expanding from (0, 0) to (1, 1)
    x = np.linspace(0.0, 1.0, 11)
    y = np.linspace(0.0, 1.0, 11)

    xy = np.array(np.meshgrid(x, y)).T.reshape(-1, 2)

    # adding a constant Z field (z = 0) to all points in the grid
    z = np.expand_dims(np.zeros(121), axis=1)

    cloud = np.append(xy, z, axis=1)

    # Adding noise/outlier: point outside the grid, at z = 2
    z_outlier = np.array([[0.5, 0.5, 2]])
    noisy_cloud = np.append(cloud, z_outlier, axis=0)

    # Expect is the regular grid
    expect = np.append(xy, z, axis=1)

    # Output should be the noisy cloud minus the outlier point --> the regular grid
    output = dm.clean_ground(noisy_cloud)

    np.testing.assert_array_almost_equal(output, expect)


def test_clean_cloth_with_noise():
    # Regular grid of points expanding from (0, 0) to (1, 1)
    x = np.linspace(0.0, 1.0, 11)
    y = np.linspace(0.0, 1.0, 11)

    xy = np.array(np.meshgrid(x, y)).T.reshape(-1, 2)

    # adding a constant Z field (z = 0) to all points in the grid
    z = np.expand_dims(np.zeros(121), axis=1)

    cloud = np.append(xy, z, axis=1)

    # Adding noise/outlier: point outside the grid, at z = 2
    z_outlier = np.array([[0.5, 0.5, 2]])

    noisy_cloud = np.append(cloud, z_outlier, axis=0)

    # Expect is the regular grid
    expect = np.append(xy, z, axis=1)

    # Output should be the noisy cloud minus the outlier point --> the regular grid
    output = dm.clean_cloth(noisy_cloud)

    np.testing.assert_array_almost_equal(output, expect)


def test_clean_cloth_already_is_clean():
    # Regular grid of points expanding from (0, 0) to (1, 1)
    x = np.linspace(0.0, 1.0, 11)
    y = np.linspace(0.0, 1.0, 11)

    xy = np.array(np.meshgrid(x, y)).T.reshape(-1, 2)

    # adding a constant Z field (z = 0) to all points in the grid
    z = np.expand_dims(np.zeros(121), axis=1)

    cloud = np.append(xy, z, axis=1)

    # Expect is the regular grid
    expect = np.append(xy, z, axis=1)

    # Output should be the regular grid --> no points removed
    output = dm.clean_cloth(cloud)

    np.testing.assert_array_almost_equal(output, expect)


def test_clean_cloth_small_cloth():
    # Regular grid of points expanding from (0, 0) to (1, 1) (ONLY 9 points)
    x = np.linspace(0.0, 1.0, 3)
    y = np.linspace(0.0, 1.0, 3)

    xy = np.array(np.meshgrid(x, y)).T.reshape(-1, 2)

    # adding a constant Z field (z = 0) to all points in the grid
    z = np.expand_dims(np.zeros(9), axis=1)

    cloud = np.append(xy, z, axis=1)

    # function should not work as it performs a 15-nearest-neighbours search.
    np.testing.assert_raises(ValueError, dm.clean_cloth, cloud)


def test_clean_cloth_barely_big_cloth():
    # Regular grid of points expanding from (0, 0) to (1, 1) (16 points)
    x = np.linspace(0.0, 1.0, 4)
    y = np.linspace(0.0, 1.0, 4)

    xy = np.array(np.meshgrid(x, y)).T.reshape(-1, 2)

    # adding a constant Z field (z = 0) to all points in the grid
    z = np.expand_dims(np.zeros(16), axis=1)

    cloud = np.append(xy, z, axis=1)

    # one points is removed, so there are 15 points
    cloud = cloud[0 : (cloud.shape[0] - 1)]

    np.testing.assert_warns(Warning, dm.clean_cloth, cloud)


def test_normalize_heights():
    # DTM:
    # Regular grid of points expanding from (0, 0) to (1, 1)
    dtm_x = np.linspace(0.0, 1.0, 11)
    dtm_y = np.linspace(0.0, 1.0, 11)
    dtm_xy = np.array(np.meshgrid(dtm_x, dtm_y)).T.reshape(-1, 2)

    # adding a constant Z field (z = 1) to all points in the grid
    dtm_z = np.expand_dims(np.zeros(121), axis=1) + 1
    dtm = np.append(dtm_xy, dtm_z, axis=1)

    # Cloud:
    cloud_x = np.expand_dims(np.linspace(0.0, 1.0, 11), axis=1)
    cloud_y = np.expand_dims(np.zeros(11), axis=1)

    cloud_xy = np.append(cloud_x, cloud_y, axis=1)

    cloud_z = np.expand_dims(np.linspace(1.0, 2.0, 11), axis=1)

    # Diagonal line from (0, 0, 1) to (1, 0, 2)
    cloud = np.append(cloud_xy, cloud_z, axis=1)

    # the normalized heights should be cloud_z - 1 for every point
    expect = np.linspace(0.0, 1.0, 11)
    output = dm.normalize_heights(cloud, dtm)

    np.testing.assert_array_almost_equal(output, expect)



__all__ = ["draw", "ground", "individualize", "primitives", "sections", "stripe"]

from .draw import *
from .ground import *
from .individualize import *
from .primitives.clustering import *
from .primitives.voxel import *
from .sections import *
from .stripe import *



### Example script ###

import os

import laspy
import numpy as np

import dendromatics as dm

### Reading the point cloud ###
os.chdir("C:/path/data")  # your path to data
filename_las = "example_data.las"  # your .las file
entr = laspy.read(filename_las)
coords = np.vstack((entr.x, entr.y, entr.z)).transpose()
# coords = np.vstack((entr.x, entr.y, entr.z, entr[field_name_z0])).transpose() for already height-normalized point clouds
# ---------------------------------------------------------#

### Normalizing the point cloud ###
clean_points = dm.clean_ground(coords)
cloth_nodes = dm.generate_dtm(clean_points)

z0_values = dm.normalize_heights(coords, cloth_nodes)
coords = np.append(coords, np.expand_dims(z0_values, axis=1), 1)

# OPTIONAL: Reducing noise from the dtm and saving it #
dtm = dm.clean_cloth(cloth_nodes)

las_dtm_points = laspy.create(point_format=2, file_version="1.2")
las_dtm_points.x = dtm[:, 0]
las_dtm_points.y = dtm[:, 1]
las_dtm_points.z = dtm[:, 2]
las_dtm_points.write(filename_las[:, -4] + "_dtm_points.las")
# ---------------------------------------------------------#

### Extracting the stripe and peeling the stems ###
upper_limit = 2.5
lower_limit = 0.5

stripe = coords[(coords[:, 3] > lower_limit) & (coords[:, 3] < upper_limit)]
stripe_stems = dm.verticality_clustering(stripe)
# ---------------------------------------------------------#

### Individualizing the trees ###
assigned_cloud, tree_vector, tree_heights = dm.individualize_trees(coords, stripe_stems)

# Saving the assigned cloud, the stems and the tree heights

# Assigned cloud
entr.add_extra_dim(laspy.ExtraBytesParams(name="dist_axes", type=np.float64))
entr.add_extra_dim(laspy.ExtraBytesParams(name="tree_ID", type=np.int32))
entr.dist_axes = assigned_cloud[:, 5]
entr.tree_ID = assigned_cloud[:, 4]
entr.write(filename_las[:, -4] + "_tree_ID_dist_axes.las")

# Stems
las_stripe = laspy.create(point_format=2, file_version="1.2")
las_stripe.x = stripe_stems[:, 0]
las_stripe.y = stripe_stems[:, 1]
las_stripe.z = stripe_stems[:, 2]

las_stripe.add_extra_dim(laspy.ExtraBytesParams(name="tree_ID", type=np.int32))
las_stripe.tree_ID = stripe_stems[:, -1]
las_stripe.write(filename_las[:, -4] + "_stripe.las")

# Tree heights
las_tree_heights = laspy.create(point_format=2, file_version="1.2")
las_tree_heights.x = tree_heights[:, 0]  # x
las_tree_heights.y = tree_heights[:, 1]  # y
las_tree_heights.z = tree_heights[:, 2]  # z
las_tree_heights.add_extra_dim(laspy.ExtraBytesParams(name="z0", type=np.int32))
las_tree_heights.z0 = tree_heights[:, 3]  # z0
las_tree_heights.add_extra_dim(laspy.ExtraBytesParams(name="deviated", type=np.int32))
las_tree_heights.deviated = tree_heights[:, 4]  # vertical deviation binary indicator
las_tree_heights.write(filename_las[:, -4] + "_tree_heights.las")
# ---------------------------------------------------------#

### Extracting and curating stems ###

expected_R = 0.5
min_h = 0.5
max_h = 25
section_width = 0.02

xyz0_coords = assigned_cloud[
    (assigned_cloud[:, 5] < expected_R)
    & (assigned_cloud[:, 3] > min_h)
    & (assigned_cloud[:, 3] < max_h + section_width),
    :,
]

stems = dm.verticality_clustering(xyz0_coords)[:, 0:6]
# ---------------------------------------------------------#

### Computing sections ###
section_len = 0.2
sections = np.arange(min_h, max_h, section_len)  # Range of uniformly spaced values within the specified interval

X_c, Y_c, R, check_circle, second_time, sector_perct, n_points_in = dm.compute_sections(stems, sections)
# ---------------------------------------------------------#

### Tilt detection ###
outliers = dm.tilt_detection(X_c, Y_c, R, sections)
# ---------------------------------------------------------#

### Drawing sections and axes ###
dm.draw_circles(
    X_c,
    Y_c,
    R,
    sections,
    check_circle,
    sector_perct,
    n_points_in,
    tree_vector,
    outliers,
    filename_las,
)
dm.draw_axes(tree_vector, filename_las)
# ---------------------------------------------------------#

### Computing DBH and tree locators
dbh_values, tree_locations = dm.tree_locator(sections, X_c, Y_c, tree_vector, sector_perct, R, outliers, n_points_in)

las_tree_locations = laspy.create(point_format=2, file_version="1.2")
las_tree_locations.x = tree_locations[:, 0]
las_tree_locations.y = tree_locations[:, 1]
las_tree_locations.z = tree_locations[:, 2]

las_tree_locations.write(filename_las[:, -4] + "_tree_locator.las")
# ---------------------------------------------------------#



#### IMPORTS ####
import timeit

import numpy as np

# -----------------------------------------------------------------------------------------------------------------------------------
# voxelate
# ----------------------------------------------------------------------------------------------------------------------------------------


def voxelate(
    cloud,
    resolution_xy,
    resolution_z,
    n_digits=5,
    X_field=0,
    Y_field=1,
    Z_field=2,
    with_n_points=True,
    verbose=True,
):
    """Function used to voxelate point clouds. It allows to use a different
    resolution for (z), but (x, y) will always share the same resolution. It
    also allows to revert the process, by creating a unique code for each point
    in the point cloud, thus voxelated cloud can be seamlessly reverted to the
    original point cloud. If `dendroptimized` is installed the optimized version
    is used instead of this one.

    Parameters
    ----------
    cloud : numpy.ndarray
        The point cloud to be voxelated. It is expected to have X, Y, Z fields.
        3D or higher array containing data with `float` type.
    resolution_xy : float
        (x, y) voxel resolution.
    resolution_z : float
        (z) voxel resolution.
    n_digits : int
        Number of digits dedicated to each coordinate ((x), (y) or (z)) during
        the generation of each point code. Defaults to 5.
    X_field : int
        Index at which (x) coordinate is stored. Defaults to 0.
    Y_field : int
        Index at which (y) coordinate is stored. Defaults to 1.
    Z_field : int
        Index at which (z) coordinate is stored. Defaults to 2.
    with_n_points : boolean
        If True, output voxelated cloud will have a field including the number
        of points that each voxel contains. Defaults to True.

    Returns
    -------
    voxelated_cloud : numpy.ndarray
        The voxelated cloud. It consists of 3 columns, each with (x), (y) and
        (z) coordinates, and an optional 4th column having the number of points
        included in each voxel if with_n_points = True.
    vox_to_cloud_ind : numpy.ndarray
        Vector containing the indexes to revert to the original point cloud
        from the voxelated cloud.
    cloud_to_vox_ind : numpy.ndarray
        Vector containing the indexes to directly go from the original point
        cloud to the voxelated cloud.
    """

    t = timeit.default_timer()

    # The coordinate minima
    cloud_min = np.min(cloud[:, [X_field, Y_field, Z_field]], axis=0)

    # Substraction of the coordinates
    cloud[:, X_field] = cloud[:, X_field] - cloud_min[0]
    cloud[:, Y_field] = cloud[:, Y_field] - cloud_min[1]
    cloud[:, Z_field] = cloud[:, Z_field] - cloud_min[2]

    if verbose:
        elapsed = timeit.default_timer() - t
        print("      -Voxelization")
        print(
            "        ",
            "Voxel resolution:",
            "{:.2f}".format(resolution_xy),
            "x",
            "{:.2f}".format(resolution_xy),
            "x",
            "{:.2f}".format(resolution_z),
            "m",
        )
        print("        ", "%.2f" % elapsed, "s: scaling and translating")

    # Generation of 'pixel code'. It provides each point with an unique identifier.
    code = (
        np.floor(cloud[:, Z_field] / resolution_z) * 10 ** (n_digits * 2)
        + np.floor(cloud[:, Y_field] / resolution_xy) * 10**n_digits
        + np.floor(cloud[:, X_field] / resolution_xy)
    )

    if not verbose:
        elapsed = timeit.default_timer() - t
        print("        ", "%.2f" % elapsed, "s: encoding")

    # Vector that contains the ordered code. It will be used to sort the code to
    # then sort the cloud.
    vox_order_ind = np.argsort(code)

    if not verbose:
        elapsed = timeit.default_timer() - t
        print("        ", "%.2f" % elapsed, "s: 1st sorting")

    # Vector that contains the indexes of said code. It will be used to restore
    # the order of points within the original cloud.
    vox_order_ind_inverse = np.argsort(vox_order_ind)

    # Sorted code.
    code = code[vox_order_ind]

    if not verbose:
        elapsed = timeit.default_timer() - t
        print("        ", "%.2f" % elapsed, "s: 2nd sorting")

    # Unique values of said 'pixel code':
    # unique_code: Unique values. They contain codified coordinates of which will
    #   later be the voxel centroids.
    # vox_first_point_id. Index corresponding to the point of each voxel that
    #   corresponds to the first point, among those in the same voxel, in the original cloud.
    # inverse_id: Indexes that allow to revert the voxelization.
    # vox_points: Number of points in each voxel
    unique_code, vox_first_point_id, inverse_id, vox_points = np.unique(
        code, return_index=True, return_inverse=True, return_counts=True
    )

    if not verbose:
        elapsed = timeit.default_timer() - t
        print("        ", "%.2f" % elapsed, "s: extracting uniques values")

    # Indexes that directly associate each voxel to its corresponding points in
    # the original cloud (unordered)
    vox_to_cloud_ind = inverse_id[vox_order_ind_inverse]

    # Indexes that directly associate each point in the original, unordered cloud
    # to its corresponding voxel
    cloud_to_vox_ind = vox_order_ind[vox_first_point_id]

    # Empty array to be filled with voxel coordinates
    voxelated_cloud = np.zeros((np.size(unique_code, 0), 3))

    # Each coordinate 'pixel code'. They will then be transformed into coordinates
    z_code = np.floor(unique_code / 10 ** (n_digits * 2))
    y_code = np.floor((unique_code - z_code * 10 ** (n_digits * 2)) / 10**n_digits)
    x_code = unique_code - z_code * 10 ** (n_digits * 2) - y_code * 10**n_digits

    voxelated_cloud[:, 0] = x_code
    voxelated_cloud[:, 1] = y_code
    voxelated_cloud[:, 2] = z_code

    if not verbose:
        elapsed = timeit.default_timer() - t
        print("        ", "%.2f" % elapsed, "s: decomposing code")

    # Transformation of x, z, y codes into X, Y, X voxel coordinates, by scaling,
    # translating and centering.
    voxelated_cloud[:, 2] = voxelated_cloud[:, 2] * resolution_z + cloud_min[2] + resolution_z / 2
    voxelated_cloud[:, 1] = voxelated_cloud[:, 1] * resolution_xy + cloud_min[1] + resolution_xy / 2
    voxelated_cloud[:, 0] = voxelated_cloud[:, 0] * resolution_xy + cloud_min[0] + resolution_xy / 2

    # Boolean parameter that includes or not a 4th column with the number of
    # points in each voxel
    if with_n_points is True:
        voxelated_cloud = np.append(voxelated_cloud, vox_points[:, np.newaxis], axis=1)

    if not verbose:
        elapsed = timeit.default_timer() - t
        print("        ", "%.2f" % elapsed, "s: rescaling and translating back")
        print(
            "        ",
            "{:.2f}".format(vox_to_cloud_ind.shape[0] / 1000000),
            "million points ->",
            "{:.2f}".format(cloud_to_vox_ind.shape[0] / 1000000),
            "million voxels",
        )
        print(
            "        ",
            "Voxels account for",
            "{:.2f}".format(cloud_to_vox_ind.shape[0] * 100 / vox_to_cloud_ind.shape[0]),
            "% of original points",
        )

    cloud[:, X_field] = cloud[:, X_field] + cloud_min[0]
    cloud[:, Y_field] = cloud[:, Y_field] + cloud_min[1]
    cloud[:, Z_field] = cloud[:, Z_field] + cloud_min[2]

    return voxelated_cloud, vox_to_cloud_ind, cloud_to_vox_ind


# monkey patch voxelization if possible
try:
    import dendroptimized

    voxelate = dendroptimized.voxelize
except ImportError:
    pass



import laspy
import numpy as np

# -----------------------------------------------------------------------------
# draw_circles
# -----------------------------------------------------------------------------


def generate_circles_cloud(
    X_c,
    Y_c,
    R,
    sections,
    check_circle,
    sector_perct,
    n_points_in,
    tree_vector,
    outliers,
    R_min=0.03,
    R_max=0.5,
    threshold=5,
    n_sectors=16,
    min_n_sectors=9,
    circa_points=200,
):
    """This function generates points that comprise the circles computed by
    fit_circle_check function, so sections can be visualized. The circles
    points cloud along with their associated meta data are returned as a Matrix
    (numpy.ndarray)

    Parameters
    ----------
    X_c : numpy.ndarray
        Matrix containing (x) coordinates of the center of the sections.
    Y_c : numpy.ndarray
        Matrix containing (y) coordinates of the center of the sections.
    R : numpy.ndarray
        Vector containing section radia.
    sections : numpy.ndarray
        Vector containing section heights (normalized heights).
    section_perct : numpy.ndarray
        Matrix containing the percentage of occupied sectors.
    n_points_in : numpy.ndarray
        Matrix containing the number of points in the inner circumferences.
    tree_vector : numpy.ndarray
        detected_trees output from individualize_trees.
    outliers : numpy.ndarray
        Vector containing the 'outlier probability' of each section.
    R_min : float
        Refer to fit_circle_check in 'sections' module. Defaults to 0.03.
    R_max : float
        Refer to fit_circle_check in 'sections' module. Defaults to 0.5.
    threshold : float
        Refer to fit_circle_check in 'sections' module. Defaults to 5.
    n_sectors : int
        Refer to fit_circle_check in 'sections' module. Defaults to 16.
    min_n_sectors: int
        Refer to fit_circle_check in sections module. Defaults to 9.
    circa_points : int
        Number of points used to draw each circle. Defaults to 200.

    Returns
    -------
    coords : numpy.ndarray
        Matrix containing the circles coordinates and their associated meta data
    """

    # Empty vector to be filled. It has as many elements as the vector containing
    # the center of the circles for a given tree.
    tree_section = X_c.shape

    # Empty array that will contain the information about each section, to then
    # be used to complete the .LAS file data.
    section_c_xyz = np.zeros([tree_section[0] * tree_section[1], 9])

    # Auxiliary index indicating which section is in use.
    section = 0

    # Double for loop to iterate through each combination of coordinates
    for i in range(tree_section[0]):
        for j in range(tree_section[1]):
            # If distance is within range (R_min, R_max), then proceed.
            if R[i, j] >= R_min and R[i, j] <= R_max:
                # Filling the array with the appropriate data
                section_c_xyz[section, :] = [
                    X_c[i, j],
                    Y_c[i, j],
                    sections[j] + tree_vector[i, 7],
                    R[i, j],
                    check_circle[i, j],
                    sector_perct[i, j],
                    n_points_in[i, j],
                    sections[j],
                    outliers[i, j],
                ]

                section = section + 1

    # Just the centers of each filled section
    centers = section_c_xyz[:section, :]

    # Number of centers
    n = centers.shape[0]

    # Empty vector to be filled with the coordinates of each circle.
    coords = np.zeros((circa_points * n, 11))

    # User-create function to tranform polar coordinates to cartesian coordinates.
    def polar_to_cart(theta, rho):
        x = rho * np.cos(theta)
        y = rho * np.sin(theta)
        return x, y

    # For loop to iterate over each circle and compute their (x, y) coordinates.
    # (z) coordinates are already given by the user.
    for i in range(n):
        start = i * circa_points
        end = (i + 1) * circa_points
        angles = np.arange(0, 2 * np.pi, 2 * np.pi / circa_points)
        radius = centers[i, 3]
        (x, y) = polar_to_cart(angles, radius)

        coords[start:end, 0] = x + centers[i, 0]  # X
        coords[start:end, 1] = y + centers[i, 1]  # Y
        coords[start:end, 2] = centers[i, 2]  # check
        coords[start:end, 3] = centers[i, 4]  # Z0
        coords[start:end, 4] = i  # Tree ID
        coords[start:end, 5] = centers[i, 5]  # sector occupancy
        coords[start:end, 6] = centers[i, 6]  # points in inner circle
        coords[start:end, 7] = centers[i, 7]  # Z0
        coords[start:end, 8] = centers[i, 3] * 2  # Diameter
        coords[start:end, 9] = centers[i, 8]  # outlier probability

        if (
            (centers[i, 5] < min_n_sectors / n_sectors * 100)
            | (centers[i, 6] > threshold)
            | (centers[i, 8] > 0.3)
            | (centers[i, 3] < R_min)
            | (centers[i, 3] > R_max)
        ):  # only happens when which_dbh == 0 # which_valid_points should be used here
            coords[start:end, 10] = 1  # does not pass quality checks
        else:
            coords[start:end, 10] = 0  # passes quality checks
    return coords


def draw_circles(
    X_c,
    Y_c,
    R,
    sections,
    check_circle,
    sector_perct,
    n_points_in,
    tree_vector,
    outliers,
    filename_las,
    R_min=0.03,
    R_max=0.5,
    threshold=5,
    n_sectors=16,
    min_n_sectors=9,
    circa_points=200,
):
    """This function generates points that comprise the circles computed by
    fit_circle_check function, so sections can be visualized. The circles are
    then saved in a LAS file, along some descriptive fields. Each circle
    corresponds on a one-to-one basis to the sections described by the user.

    Parameters
    ----------
    X_c : numpy.ndarray
        Matrix containing (x) coordinates of the center of the sections.
    Y_c : numpy.ndarray
        Matrix containing (y) coordinates of the center of the sections.
    R : numpy.ndarray
        Vector containing section radia.
    sections : numpy.ndarray
        Vector containing section heights (normalized heights).
    section_perct : numpy.ndarray
        Matrix containing the percentage of occupied sectors.
    n_points_in : numpy.ndarray
        Matrix containing the number of points in the inner circumferences.
    tree_vector : numpy.ndarray
        detected_trees output from individualize_trees.
    outliers : numpy.ndarray
        Vector containing the 'outlier probability' of each section.
    filename_las : char
        File name for the output file.
    R_min : float
        Refer to fit_circle_check in 'sections' module. Defaults to 0.03.
    R_max : float
        Refer to fit_circle_check in 'sections' module. Defaults to 0.5.
    threshold : float
        Refer to fit_circle_check in 'sections' module. Defaults to 5.
    n_sectors : int
        Refer to fit_circle_check in 'sections' module. Defaults to 16.
    min_n_sectors: int
        Refer to fit_circle_check in sections module. Defaults to 9.
    circa_points : int
        Number of points used to draw each circle. Defaults to 200.
    """
    coords = generate_circles_cloud(
        X_c,
        Y_c,
        R,
        sections,
        check_circle,
        sector_perct,
        n_points_in,
        tree_vector,
        outliers,
        R_min,
        R_max,
        threshold,
        n_sectors,
        min_n_sectors,
        circa_points,
    )

    # LAS file containing circle coordinates.
    las_circ = laspy.create(point_format=2, file_version="1.2")
    las_circ.x = coords[:, 0]
    las_circ.y = coords[:, 1]
    las_circ.z = coords[:, 2]

    # All extra fields.

    # las_circ.add_extra_dim(laspy.ExtraBytesParams(name = "check", type = np.int32))
    # las_circ.check = coords[:, 3]

    las_circ.add_extra_dim(laspy.ExtraBytesParams(name="tree_ID", type=np.int32))
    las_circ.tree_ID = coords[:, 4]

    las_circ.add_extra_dim(laspy.ExtraBytesParams(name="sector_occupancy_percent", type=np.float64))
    las_circ.sector_occupancy_percent = coords[:, 5]

    las_circ.add_extra_dim(laspy.ExtraBytesParams(name="pts_inner_circle", type=np.int32))
    las_circ.pts_inner_circle = coords[:, 6]

    las_circ.add_extra_dim(laspy.ExtraBytesParams(name="Z0", type=np.float64))
    las_circ.Z0 = coords[:, 7]

    las_circ.add_extra_dim(laspy.ExtraBytesParams(name="Diameter", type=np.float64))
    las_circ.Diameter = coords[:, 8]

    las_circ.add_extra_dim(laspy.ExtraBytesParams(name="outlier_prob", type=np.float64))
    las_circ.outlier_prob = coords[:, 9]

    las_circ.add_extra_dim(laspy.ExtraBytesParams(name="quality", type=np.int32))
    las_circ.quality = coords[:, 10]

    las_circ.write(filename_las)


# -----------------------------------------------------------------------------
# draw_axes
# -----------------------------------------------------------------------------


def generate_axis_cloud(
    tree_vector,
    line_downstep=0.5,
    line_upstep=10.0,
    stripe_lower_limit=0.5,
    stripe_upper_limit=2.5,
    point_interval=0.01,
):
    """This function generates points that comprise the axes computed by
    individualize_trees, so that they can be visualized. It output two
    numpy.ndarray that describes the point cloud of the axis
    and their associated tilt.

    Parameters
    ----------
    tree_vector : numpy.ndarray
        detected_trees output from individualize_trees.
    filename_las : char
        File name for the output file
    line_downstep : float
        From the stripe centroid, how much (downwards direction) will the drawn
        axes extend (units is meters). Defaults to 0.5.
    line_upstep : float
        From the stripe centroid, how much (upwards direction) will the drawn
        axes extend (units is meters). Defaults to 10.0.
    stripe_lower_limit : float
        Lower (vertical) limit of the stripe (units is meters). Defaults to 0.7.
    stripe_upper_limit : float
        Upper (vertical) limit of the stripe (units is meters). Defaults to 3.5.
    point_interval : float
        Step value used to draw points (unit is meters). Defaults to 0.01.

    Returns
    --------
    axes_point : numpy.ndarray
        Matrix that describes the point cloud of the axes
    tilt : numpy.ndarray
        Matrix that describes the tilt of each axes
    """
    stripe_centroid = (stripe_lower_limit + stripe_upper_limit) / 2.0
    mean_descend = stripe_centroid + line_downstep
    mean_rise = line_upstep - stripe_centroid

    up_iter = int(np.floor(mean_rise / point_interval))
    down_iter = int(mean_descend / point_interval)

    axes_points = np.zeros((tree_vector.shape[0] * (up_iter + down_iter), 3))
    tilt = np.zeros(tree_vector.shape[0] * (up_iter + down_iter))

    ind = 0
    for i in range(tree_vector.shape[0]):
        if np.sum(np.exp2(tree_vector[i, 1:4])) > 0:
            vector = -tree_vector[i, 1:4] if tree_vector[i, 3] < 0 else tree_vector[i, 1:4]
            next_ind = ind + up_iter + down_iter
            axes_points[ind:next_ind] = (
                np.column_stack(
                    (
                        np.arange(-down_iter, up_iter),
                        np.arange(-down_iter, up_iter),
                        np.arange(-down_iter, up_iter),
                    )
                )
                * vector
                * point_interval
                + tree_vector[i, 4:7]
            )
            tilt[ind:next_ind] = tree_vector[i, 8]
            ind = next_ind

    axes_points = axes_points[:ind]
    return axes_points, tilt


def draw_axes(
    tree_vector,
    filename_las,
    line_downstep=0.5,
    line_upstep=10.0,
    stripe_lower_limit=0.5,
    stripe_upper_limit=2.5,
    point_interval=0.01,
):
    """This function generates points that comprise the axes computed by
    individualize_trees, so that they can be visualized. The axes are then
    saved in a LAS file, along some descriptive fields. Each axis corresponds
    on a one-to-one basis to the individualized trees.

    Parameters
    ----------
    tree_vector : numpy.ndarray
        detected_trees output from individualize_trees.
    filename_las : char
        File name for the output file
    line_downstep : float
        From the stripe centroid, how much (downwards direction) will the drawn
        axes extend (units is meters). Defaults to 0.5.
    line_upstep : float
        From the stripe centroid, how much (upwards direction) will the drawn
        axes extend (units is meters). Defaults to 10.0.
    stripe_lower_limit : float
        Lower (vertical) limit of the stripe (units is meters). Defaults to 0.5.
    stripe_upper_limit : float
        Upper (vertical) limit of the stripe (units is meters). Defaults to 2.5.
    point_interval : float
        Step value used to draw points (unit is meters). Defaults to 0.01..
    """
    axes_points, tilt = generate_axis_cloud(
        tree_vector,
        line_downstep,
        line_upstep,
        stripe_lower_limit,
        stripe_upper_limit,
        point_interval,
    )

    las_axes = laspy.create(point_format=2, file_version="1.2")
    las_axes.x = axes_points[:, 0]
    las_axes.y = axes_points[:, 1]
    las_axes.z = axes_points[:, 2]
    las_axes.add_extra_dim(laspy.ExtraBytesParams(name="tilting_degree", type=np.float64))
    las_axes.tilting_degree = tilt

    las_axes.write(filename_las)
